import unittest, time
from kcapi.rest.crud import KeycloakCRUD
from kcapi.rest.targets import Targets
from kcapi.rest.url import RestURL
from .testbed import TestBed
import json

def load_sample(fname):
    f = open(fname)
    file1 = json.loads(f.read())
    f.close()
    return file1

def exist(that, res, name):
    ret  = res.findFirstByKV('username', name)
    that.assertEqual(name, ret['firstName'], 'We expect a role ['+name+'] to be created')
    return ret['id']



def test_complete_CRUD(that, users):
        ## POST
        state = users.create(that.USER_DATA)
        that.assertTrue(state, 'fail while posting')
     
        ret  = users.findFirstByKV('username', 'pepe')
        that.assertEqual('pepe', ret['firstName'], 'We expect a user with pepe as username to be created')
        
        ## UPDATE
        state = users.update(ret['id'], {'firstName': 'pedro'})
        that.assertTrue(state, 'fail while updating')
        
        ret  = users.findFirstByKV('firstName', 'pedro')
        that.assertTrue(ret != False, 'Something wrong updating the resource.')
        that.assertEqual('pedro', ret['firstName'], 'We expect a user with pepe as username to be created')
        
        ## GET
        usr = users.get(ret['id']).resp().json()
        that.assertEqual(ret['id'], usr['id'])
        
        _all = users.all()
        that.assertEqual(len(_all), 4, 'All users amount to one')


        ## DELETE
        remove_state = users.remove(ret['id']).isOk()
        that.assertTrue(remove_state)
        removed = users.findFirstByKV('firstName', 'pedro')

        that.assertFalse(removed)



class Testing_User_API(unittest.TestCase):

    def testing_crud_API(self):
        token = self.testbed.token

        users = KeycloakCRUD()
        users.targets = Targets.makeWithURL(str(self.USER_ENDPOINT))
        users.token = token

        test_complete_CRUD(self, users)
   
    @classmethod
    def setUpClass(self):
       
        self.testbed = TestBed()
        self.testbed.createRealms()
        self.testbed.createUsers()
        self.REALM = self.testbed.REALM
       
        self.USER_ENDPOINT = RestURL(self.testbed.ENDPOINT)
        self.USER_ENDPOINT.addResources(['auth', 'admin', 'realms', self.REALM, 'users']) 

        self.USER_DATA = {"enabled":True,"attributes":{},"username":"pepe","emailVerified":"", "firstName": 'pepe'}
       
    @classmethod
    def tearDownClass(self):
        self.testbed.goodBye()
        return True

if __name__ == '__main__':
    unittest.main()



import unittest, time, json
from kcapi import OpenID, Keycloak
from .testbed import TestBed 

WRONG_URL =  'https://sso-wrong-cvaldezr-stage.apps.sandbox-m2.ll9k.p1.openshiftapps.com'

def test_kc(that, kc):
    adm = kc.admin()
    realms = adm.all()
    that.assertTrue(len(realms) > 0, "We should get some realms from the server.")

    kc_resource= kc.build('clients', 'master')
    clients = kc_resource.all()
    that.assertTrue(len(clients) > 0, "We should get some clients from the server.")

    poly_roles = kc.build('roles', 'master')
    roles = poly_roles.all()
    that.assertTrue(len(roles) > 0, "We should get some roles from the server.")


def test_kc_crud(that, kc):
    payload = {"enabled":True,"attributes":{},"redirectUris":[],"clientId":"test-client","protocol":"openid-connect"}
    clients = kc.build('clients', that.testbed.REALM)
    that.assertTrue(clients.create(payload).isOk(), 'Client should be created')
    clients = clients.findFirstByKV('clientId', 'test-client')
    that.assertTrue(len(clients) > 0, 'We should get a client back')

class Testing_OpenID(unittest.TestCase):
    def test_oidc_ctor(self): 
        try:
            OpenID({"client_id":"admin-cli"}, WRONG_URL)
        except Exception as E:
            self.assertEqual("password" in str(E), True)

        try:
            OpenID({"client_id":"admin-cli", "realm": "my_realm", "password": "xxx"}, WRONG_URL)
        except Exception as E:
            self.assertEqual("username" in str(E), True)

        self.assertRaises(Exception, lambda: OpenID({"client_id":"admin-cli"})) 

    def test_oidc_login_on_wrong_url(self):
        oid_client = OpenID({
            "client_id": "admin-cli",
            "username": "6OPtWY33", 
            "password":"I6gglTeDLlmmpLYoAAUMcFQqNOMjw5dA", 
            "grant_type":"password",
            "realm" : "master"
            }, WRONG_URL)
        
        self.assertRaises(Exception, lambda: oid_client.getToken()) 

    def test_oidc_login_on_wrong_password(self):
        oid_client = OpenID({
            "client_id": "admin-cli",
            "username": "6OPtWY33", 
            "password":"I6gglTeDLlmmpLYoAAUMcFQqNOMjw5dA", 
            "grant_type":"password",
            "realm" : "master"
            }, self.ENDPOINT)

        try:
            oid_client.getToken()
        except Exception as E:
            self.assertEqual("Unauthorized" in str(E), True)


    def test_oidc_login(self):
        oid_client = OpenID({
            "client_id": "admin-cli",
            "username": self.testbed.USER, 
            "password": self.testbed.PASSWORD, 
            "grant_type":"password",
            "realm" : "master"
            }, self.ENDPOINT)

        token = oid_client.getToken()

        self.assertIsNotNone(token)
        self.assertTrue(len(str(token)) > 0)


    def testing_admin_instantiation(self):
        oid_client = OpenID({
            "client_id": "admin-cli",
            "username": self.testbed.USER,
            "password": self.testbed.PASSWORD,
            "grant_type": "password",
            "realm": "master"
        }, self.ENDPOINT)

        token = oid_client.getToken()
        kc = Keycloak(token, self.ENDPOINT)

        test_kc(self, kc)


    def testing_admin_instantiation_with_token_refresh(self):
        oid_client = OpenID({
            "client_id": "admin-cli",
            "username": self.testbed.USER,
            "password": self.testbed.PASSWORD,
            "grant_type": "password",
            "realm": "master"
        }, self.ENDPOINT)

        token = oid_client.getToken()

        new_kc = Keycloak.instantiate_with_refresh_token(refresh_token=str(token.refresh_token), url=self.ENDPOINT)
        test_kc(self, new_kc)

    @unittest.skip
    def testing_admin_instantiation_with_string_request(self):
        oid_client = OpenID({
            "client_id": "admin-cli",
            "username": self.testbed.USER,
            "password": self.testbed.PASSWORD,
            "grant_type": "password",
            "realm": "master"
        }, self.ENDPOINT)

        token_str = open('./test/token_example/token').read()

        json_kc = Keycloak.instantiate_with_raw_json(json_string=str(token_str), url=self.ENDPOINT, client_id='security-admin-console')
        test_kc(self, json_kc)


    def testing_token_refresh_mechanism_triggering_manual_refresh(self):
        token = OpenID.createAdminClient(self.USER, self.PASSWORD, url=self.ENDPOINT).getToken()

        kc = Keycloak.instantiate_with_raw_json(json_string=str(token), url=self.ENDPOINT)
        aa  = kc.admin()
        aa.create({"enabled": "true", "id": 'deleteme-5', "realm": "deleteme_5"})
        test_kc(self, kc)

        unique = {}
        for i in range(6):
            temp_token = token.refresh()
            self.assertTrue(not temp_token.get_token() in unique)
            unique[temp_token.get_token()] = 1
            kc_refreshed = Keycloak(temp_token, self.ENDPOINT)
            test_kc(self, kc_refreshed)

        aa.token.refresh()
        aa.removeFirstByKV('realm', 'deleteme-5')

    def testing_token_refresh_mechanism_with_timeout(self):
        str_token = str(self.kc.token)
        kc = Keycloak.instantiate_with_raw_json(url=self.testbed.ENDPOINT, json_string=str_token)
        groups = kc.build('groups', self.realm)

        state = groups.create({"name":"delete_this_group"}).isOk()
        self.assertTrue(state)

        time.sleep(65)

        _kc = Keycloak.instantiate_with_raw_json(url=self.testbed.ENDPOINT, json_string=str_token)
        group_using_expired_token =  _kc.build('groups', self.realm)
        test_kc(self, _kc)
        test_kc_crud(self, kc)

        state_after_delay = group_using_expired_token.removeFirstByKV("name", "delete_this_group")
        self.assertTrue(state_after_delay)


    @classmethod
    def setUpClass(self):
        self.testbed = TestBed()
        self.testbed.createRealms()
        self.kc = self.testbed.getKeycloak()
        self.realm = self.testbed.REALM 
        self.master_realm = self.testbed.getAdminRealm()
        self.ENDPOINT = self.testbed.ENDPOINT
        self.USER = self.testbed.USER
        self.PASSWORD = self.testbed.PASSWORD

    @classmethod
    def tearDownClass(self):
        self.testbed.goodBye()
        return True

if __name__ == '__main__':
    unittest.main()



from urllib.parse import urlparse

EMPTY_STRING = ""
DEBUG = False


class RestURL:
    def __get_res_from_url(self, uri):
        res = uri.path.split('/')
        res = filter(lambda re: re != "", res)

        return list(res)

    def host_url(self):
        default_scheme = 'https' if not self.uri.scheme else self.uri.scheme

        return default_scheme + "://" + self.uri.hostname

    def __get_target(self):
        return self.host_url() if not self.uri.port else self.host_url() + ":" + str(self.uri.port)

    def __init__(self, url, resources=[]):
        if not url:
            raise Exception("Missing URL.")

        if 'http' not in url:
            raise Exception("No HTTP schema found: <http(s)>://host")

        self.uri = urlparse(url)
        self.url = url
        self.resources = self.__get_res_from_url(self.uri)
        self.resources = self.resources + resources

    def pushResource(self, fragment):
        if not self.resources:
            self.resources.append(fragment)
            return True

        if self.resources[-1] != fragment:
            self.resources.append(fragment)
            return True

        return False

    def setId(self, _id):
        self.addResource(_id)

    def addResources(self, resources):
        for resource in resources:
            if resource:
                self.pushResource(resource)

        return self

    def addResource(self, name=None, value=None):
        if name:
            self.pushResource(name)
        if value:
            self.pushResource(value)
        return self

    def removeResources(self, listOfResources):
        for resource in listOfResources:
            if resource in self.resources:
                index = self.resources.index(resource)
                self.resources.pop(index)

    def removeLast(self):
        if len(self.resources) > 0:
            self.resources.pop()
        return self

    def getCurrentResource(self):
        if not self.resources:
            raise Exception("No URL resources found.")
        return self.resources[-1]

    def replaceResource(self, origin, replacement):
        self.resources = [replacement if entry == origin else entry for entry in self.resources]

    def replaceCurrentResourceTarget(self, resReplacementName):
        popped = self.resources.pop()
        self.resources.append(resReplacementName)
        return self

    def buildResURL(self):
        resource_section = ""
        if len(self.resources) == 0:
            return EMPTY_STRING

        for part in self.resources:
            resource_section = resource_section + "/" + str(part)

        return resource_section

    def __repr__(self):
        return self.__get_target() + self.buildResURL()

    def copy(self):
        return RestURL(self.__get_target(), self.resources.copy())

    def target(self):
        return self.__get_target()



import unittest, time
from .testbed import TestBed
import json


def load_sample(file_name):
    f = open(file_name)
    payload = json.loads(f.read())
    f.close()

    return payload


class TestingRealmAPI(unittest.TestCase):

    def testing_realm_api_methods(self):
        realms = self.testbed.getKeycloak().build('realms', self.REALM)
        self.assertTrue(hasattr(realms, 'caches'))


    def testing_realm_cache_reset(self):
        realms = self.testbed.getKeycloak().build('realms', self.REALM)

        caches = realms.caches(self.REALM)

        self.assertEqual(caches.clearRealmCache().resp().status_code, 204)

    def testing_user_cache_reset(self):
        realms = self.testbed.getKeycloak().build('realms', self.REALM)

        caches = realms.caches(self.REALM)
        self.assertEqual(caches.clearUserCache().resp().status_code, 204)

    def testing_key_cache_reset(self):
        realms = self.testbed.getKeycloak().build('realms', self.REALM)

        caches = realms.caches(self.REALM)

        self.assertEqual(caches.clearKeyCache().resp().status_code, 204)

    def testing_complex_realm_publishing(self):
        admin = self.testbed.getAdminRealm()
        realm_cfg = load_sample('./test/payloads/complex_realms.json')
        creation_state = admin.create(realm_cfg).isOk()
        self.assertTrue(creation_state, 'This realm should be created')



    @classmethod
    def setUpClass(self):
        self.testbed = TestBed()
        self.testbed.createRealms()
        self.testbed.createUsers()
        self.testbed.createClients()


        self.REALM = self.testbed.REALM

    @classmethod
    def tearDownClass(self):
        self.testbed.goodBye()
        admin = self.testbed.getAdminRealm()
        realm_cfg = load_sample('./test/payloads/complex_realms.json')
        admin.removeFirstByKV("realm", realm_cfg["realm"], custom_key="realm")


if __name__ == '__main__':
    unittest.main()



from .crud import KeycloakCRUD

def hack_rest_roles_remove_endpoint(that, kc):
    custom_delete = that.targets.targets['delete'].copy()
    custom_delete.replaceResource('clients', 'roles-by-id')
    kc.targets.targets['delete'] = custom_delete

    return kc

class Clients(KeycloakCRUD):
    def roles(self, client_query):
        client_id = super().findFirst(client_query)['id']
        child = KeycloakCRUD.get_child(self, client_id, 'roles')
        return hack_rest_roles_remove_endpoint(self, child)



import unittest, time
from kcapi import OpenID, Keycloak
from .testbed import TestBed 

ADMIN_USER = "admin"
ADMIN_PSW  = "admin1234"
REALM = "test_heroes_test"
ENDPOINT = 'https://sso-cvaldezr-stage.apps.sandbox-m2.ll9k.p1.openshiftapps.com'


TEST_REALM = "TESTING"

class Testing_SSO_API(unittest.TestCase):
    def testing_CRUD_api(self):
        realm = self.master_realm

        resp = realm.create({"enabled": "true", "id": TEST_REALM, "realm": TEST_REALM})
        self.assertEqual(resp.isOk(), True)

        findings = realm.get(TEST_REALM).verify().resp().json()
        self.assertEqual(findings['id'], TEST_REALM)

        obj = {'displayName':"MyRealm" }
        state_update = realm.update(TEST_REALM, obj).isOk()
        self.assertEqual(state_update, True)

        updated_object = realm.get(TEST_REALM).verify().resp().json()
        self.assertEqual(updated_object['displayName'], "MyRealm")

        exists = realm.exist(TEST_REALM) 
        self.assertTrue(exists)

        do_not_exists = realm.exist("DoesntExist") 
        self.assertFalse(do_not_exists)

        remove_state = realm.remove(TEST_REALM).isOk()
        self.assertEqual(remove_state, True)

    def testing_findFirstByKV(self): 
        users = self.kc.build("users", self.realm) 
        batman = users.findFirstByKV('firstName', 'Bruce')
        self.assertEqual(batman['username'], 'batman')

    def testing_updateUsingKV(self): 
        users = self.kc.build("users", self.realm) 
        status = users.updateUsingKV('firstName', 'Bruce', {'firstName': 'Bruno', 'lastName': 'Diaz'})
        self.assertTrue(status)
        batman = users.findFirstByKV('firstName', 'Bruno')
        self.assertEqual(batman['username'], 'batman')
        self.assertEqual(batman['firstName'], 'Bruno')
        self.assertEqual(batman['lastName'], 'Diaz')

    def testing_existByKV(self): 
        users = self.kc.build("users", self.realm) 
        batman = users.existByKV('firstName', 'Bruce')
        self.assertEqual(batman, True )

    def testing_removeFirstByKV(self):
        users = self.kc.build("users", self.realm) 
        all_users = users.findAll().verify().resp().json()

        self.assertEqual(len(all_users), 3)
        aqua = users.removeFirstByKV('username', 'aquaman')
        self.assertTrue(aqua)

        all_users = users.findAll().verify().resp().json()
        self.assertEqual(len(all_users), 2)
        for usr in all_users: 
            self.assertTrue( usr['username'] in ['batman', 'superman'] )

        

    def testing_client_API(self): 
        client_payload = {"enabled":True,"attributes":{},"redirectUris":[],"clientId":"deleteme","protocol":"openid-connect"}

        clients = self.kc.build("clients", self.realm)
        state = clients.create(client_payload).isOk()
        self.assertEqual(state, True)

    def testing_client_API(self): 
        clientAPI = self.kc.build("clients", self.realm)
        state = clientAPI.findAll().verify().resp()
        self.assertEqual(state.status_code, 200)
     
    
    def testing_with_no_token(self): 
        try:
            self.kc = Keycloak(None, ENDPOINT)
            self.assertTrue(False)
        except Exception as E: 
            self.assertEqual("No authentication token provided" in str(E), True)

        try:
            self.kc = Keycloak('placeholder', None)
            self.assertTrue(False)
        except Exception as E: 
            self.assertEqual("No Keycloak endpoint URL" in str(E), True)

    def testing_roles_creation_and_removal(self):
        role = {"name": "magic"}
        roles = self.kc.build("roles", self.realm)
        state = roles.create(role).isOk() 
        self.assertTrue(state)

        ammount_of_roles = len(roles.findAll().resp().json())

        self.assertEqual(ammount_of_roles, 3)

        self.assertTrue(roles.removeFirstByKV("name", "magic"))
        ammount_of_roles = len(roles.findAll().resp().json())
        self.assertEqual(ammount_of_roles, 2)

    def testing_case_sensitive_resource(self):
        myCaseTrickyUser = {"enabled":'true',"attributes":{},"username":"The Punisher","firstName":"Bruce", "lastName":"Wayne", "emailVerified":""}

        users = self.kc.build('users', self.realm) 
        state = users.create(myCaseTrickyUser).isOk()
        self.assertTrue(state)

        removed = users.removeFirstByKV('username', 'The Punisher')
        self.assertTrue(removed)

        is_empty = not users.findFirstByKV('username', 'The Punisher')
        self.assertTrue(is_empty)

    @classmethod
    def setUpClass(self):
        self.testbed = TestBed(REALM, ADMIN_USER, ADMIN_PSW, ENDPOINT)
        self.testbed.createRealms()
        self.testbed.createUsers()
        self.testbed.createClients()
        self.kc = self.testbed.getKeycloak()
        self.realm = self.testbed.REALM 
        self.master_realm = self.testbed.getAdminRealm()
        
    @classmethod
    def tearDownClass(self):
        #self.testbed.goodBye()
        if self.master_realm.exist(TEST_REALM): 
            #self.master_realm.remove(TEST_REALM)
            return True
        return True

if __name__ == '__main__':
    unittest.main()



from .crud import KeycloakCRUD
from .helper import ValidateParams


class Users(KeycloakCRUD):
    def __lazy_load_groups(self):
        groups = KeycloakCRUD() 
        groups.token = self.token
        groups.targets = self.targets.copy()
        groups.targets.change('groups')

        return groups

    def __findGroup(self, group): 
        groups = self.__lazy_load_groups()
        return groups.findFirst(group)

    def __userGroupMappingAPI(self, userID): 
        kc = KeycloakCRUD() 
        kc.token = self.token
        kc.targets = self.targets.copy()

        kc.targets.addResources([userID, 'groups'])
        return kc

    def joinGroup(self, user, group): 
        userID = self.findFirst(user)['id']
        groupID = self.__findGroup(group)['id']

        requestBody = {'groupId': groupID, 'userId': userID} 

        return self.__userGroupMappingAPI(userID).update(groupID, requestBody)

    def groups(self, user):
        userID = super().findFirst(user)['id']
        return self.__userGroupMappingAPI(userID).all()

    def leaveGroup(self, user, group):
        userID  = super().findFirst(user)['id']
        groupID = self.__findGroup(group)['id']

        return self.__userGroupMappingAPI(userID).remove(groupID)

    def groupMapping(self, user):
        userID = self.findFirst(user)['id']
        return self.__userGroupMappingAPI(userID)

    # 
    # credentials: {type: "password", value: "passphrases", temporary: true} 
    # type: password is the credential type supported by Keycloak.
    # value: Here we put the passphrase (required) 
    # temporary: **true** Means that this password would works the first time but it will force the user to setup a new one. 
    def updateCredentials(self, user, credentials):

        params = {"type": "password", "temporary":True}
        userID = super().findFirst(user)['id']
        
        params.update(credentials)
        ValidateParams(['type', 'value', 'temporary'],params)
        
        credentials = KeycloakCRUD()
        credentials.token = self.token 
        credentials.targets = self.targets.addResourcesFor('update', [userID, 'reset-password'])

        return credentials.update('', params)





import json
import unittest

from kcapi.rest.targets import Targets

TEST_URL = 'https://hello.com'

def load_sample(fname):
    f = open(fname)
    file1 = json.loads(f.read())
    f.close()
    return file1

def exist(that, res, name):
    ret  = res.findFirstByKV('username', name)
    that.assertEqual(name, ret['firstName'], 'We expect a role ['+name+'] to be created')
    return ret['id']


class Testing_User_API(unittest.TestCase):

    
    def testing_targets_class(self):
        target = Targets()
        target.makeWithURL(TEST_URL)
        

    def testing_targets_class_with_wrong_method(self):
        target = Targets.makeWithURL(TEST_URL)
        
        try:
            target.url('wrong') 
        except Exception as E:
            self.assertEqual(str(E), 'Method: wrong not supported')

    def testing_add_resources(self):
        target = Targets.makeWithURL(TEST_URL)
        target.addResourcesFor('create', ['bye']) 
        self.assertEqual('https://hello.com/bye', str(target.url('create')))
        self.assertEqual('https://hello.com', str(target.url('delete')))
        self.assertEqual('https://hello.com', str(target.url('read')))


    def testing_targets_copy_method(self):
        target = Targets.makeWithURL(TEST_URL)
        t2 = target.copy()

        t2.addResourcesFor('create', ['bye']) 
        self.assertNotEqual(str(t2.url('create')), t2.url('create'))
        self.assertEqual('https://hello.com/bye', str(t2.url('create')))
        self.assertEqual('https://hello.com', str(target.url('create')))



    def testing_targets_remove_last_method(self):
        target = Targets.makeWithURL(TEST_URL)
        target.addResourcesFor('create', ['bye']) 
        target.removeLast() 
        self.assertEqual('https://hello.com', str(target.url('create')))
        self.assertEqual('https://hello.com', str(target.url('delete')))
        self.assertEqual('https://hello.com', str(target.url('read')))

    def testing_targets_change_method(self):
        target = Targets.makeWithURL(TEST_URL)
        target.addResources(['bye']) 
        target.change('hello') 

        self.assertEqual('https://hello.com/hello', str(target.url('create')))
        self.assertEqual('https://hello.com/hello', str(target.url('delete')))
        self.assertEqual('https://hello.com/hello', str(target.url('read')))


    def testing_remove_resources_method(self):
        target = Targets.makeWithURL(TEST_URL)
        target.addResources(['1','2','hh','bye']) 
        target.removeResources(['1', '2', 'hh'])

        self.assertEqual('https://hello.com/bye', str(target.url('create')))
        self.assertEqual('https://hello.com/bye', str(target.url('read')))
        self.assertEqual('https://hello.com/bye', str(target.url('delete')))
        self.assertEqual('https://hello.com/bye', str(target.url('update')))


    def testing_replace_resources_for_method(self):
        target = Targets.makeWithURL(TEST_URL)
        target.addResources(['1','2','hh','bye']) 
        target.removeResources(['1', '2', 'hh'])

        self.assertEqual('https://hello.com/bye', str(target.url('create')))
        self.assertEqual('https://hello.com/bye', str(target.url('read')))
        self.assertEqual('https://hello.com/bye', str(target.url('delete')))
        self.assertEqual('https://hello.com/bye', str(target.url('update')))





        
            
if __name__ == '__main__':
    unittest.main()



from .targets import Targets


def IdentityProviderURLBuilder(url):
    targets = Targets.makeWithURL(url)
    targets.removeLast()
    targets.addResources(['identity-provider' , 'instances'])

    return targets



def add_failure_recovery(kc, fn):
    retry_count = 0
    def recovery(*args):
        resp = fn(*args)
        if resp.response.status_code == 401:
            kc.token = kc.token.refresh()
            print('Refresh!!!')
            return fn(*args)
        return resp

    return recovery

def add_failiure_recovery_decorator(kc):
    kc.create = add_failure_recovery(kc, kc.create)
    kc.update = add_failure_recovery(kc, kc.update)
    kc.findAll = add_failure_recovery(kc, kc.findAll)
    kc.get = add_failure_recovery(kc, kc.get)

    return kc


import unittest
import json

from .testbed import TestBed


def load_sample(fname):
    f = open(fname)
    file1 = json.loads(f.read())
    f.close()
    return file1


class TestClients(unittest.TestCase):
    def test_client_extra_methods(self):
        clients = self.testbed.getKeycloak().build('clients', self.REALM)
        self.assertIsNotNone(clients.roles, 'The client object should have a roles method.')

    def test_client_roles(self):
        client_payload = load_sample('./test/payloads/client.json')
        clients = self.testbed.getKeycloak().build('clients', self.REALM)
        svc_state = clients.create(client_payload).isOk()
        self.assertTrue(svc_state, 'The service should return a 200.')

        ret = clients.findFirstByKV('clientId', client_payload['clientId'])
        self.assertNotEqual(ret, [], 'It should return the posted client')

        client_query = {'key': 'clientId', 'value': client_payload['clientId']}
        client_roles_api = clients.roles(client_query)
        svc_roles_state = client_roles_api.create(
            {"name": "new-role", "description": "here should go a description."}).isOk()
        self.assertTrue(svc_roles_state, 'The client_roles service should return a 200.')

        ret_client_roles = client_roles_api.findFirstByKV('name', 'new-role')
        self.assertNotEqual(ret_client_roles, [], 'It should return the posted client')


    def test_client_roles_removal(self):
        client_payload = load_sample('./test/payloads/client.json')
        client_payload['clientId'] = 'test_client_roles_removal'
        client_role_name = "deleteme-role"

        clients = self.testbed.getKeycloak().build('clients', self.REALM)
        svc_state = clients.create(client_payload).isOk()

        client_query = {'key': 'clientId', 'value': client_payload['clientId']}
        client_roles_api = clients.roles(client_query)
        svc_roles_state = client_roles_api.create(
            {"name": client_role_name, "description": "A role that need to be deleted."}).isOk()

        ret_client_roles = client_roles_api.findFirstByKV('name', client_role_name)
        self.assertNotEqual(ret_client_roles, [], 'It should return the posted client')

        result = client_roles_api.removeFirstByKV('name', client_role_name)
        self.assertTrue(result, 'The server should return ok.')

        self.assertEqual(client_roles_api.findFirstByKV('name', client_role_name), [], 'It should return the posted client')

    @classmethod
    def setUpClass(self):
        self.testbed = TestBed()
        self.testbed.createRealms()
        self.testbed.createUsers()
        self.testbed.createClients()
        self.REALM = self.testbed.REALM

    @classmethod
    def tearDownClass(self):
        self.testbed.goodBye()
        return 1



class AuthFlowsPublisher(): 
    def __init__(self, authenticationFlowAPI): 
        self.flowAPI = authenticationFlowAPI

    def __flow(self, flow):
        alias = flow['displayName'] 
        pid = None if not 'provider' in flow else flow['providerId'] 
        provider = 'registration-page-flow' if not pid else pid 
        flow_type = 'basic-flow' if not pid else 'form-flow' 

        # WARN: The value description is not well validated in Keycloak, it can return 500.
        return {
                "alias":alias,
                "type": flow_type,
                "description":"empty", 
                "provider": provider,
        }

    def __executions(self, execution):
        provider = execution['providerId'] 
        return {'provider': provider} 

    def __isAuthFlow(self, body): 
        return 'authenticationFlow' in body 

    def __getNodeLevel(self, node):
        return node['level']

    def __updateOtherFields(self, parent, flow): 
        publishedFlows = parent.all()

        for publishedFlow in publishedFlows: 
            displayIsEqual = publishedFlow['displayName'] == flow['displayName']
            levelIsEqual = publishedFlow['level'] == flow['level']
            indexIsEqual = publishedFlow['index'] == flow['index']

            if displayIsEqual and levelIsEqual and indexIsEqual: 
                publishedFlow['requirement'] = flow['requirement']
                parent.update(obj=publishedFlow).isOk()
                return True


        return False


    def publish(self, rootNode, flows):  
        if not isinstance(flows, list):  
            raise Exception("Bad Parameters for Authentication Flow: auth_flow parameter should be an array.")

        rootNode = rootNode    
        nodes = { 0: rootNode } 
        current_level = 0

        rootFlow = self.flowAPI.executions(rootNode)
        
        for flow in flows:
            current_level = self.__getNodeLevel(flow)
            parent = nodes[current_level]

            if self.__isAuthFlow(flow):
               authenticationFlow = self.__flow(flow) 
               nodes[current_level + 1] = authenticationFlow
               self.flowAPI.flows(parent).create(authenticationFlow).isOk()
            else:
               execution = self.__executions(flow)
               self.flowAPI.executions(parent).create(execution).isOk()

        
            if not self.__updateOtherFields(rootFlow, flow):
                raise Exception('There is an inconsistency problem: Changes are not taking place in the server, latency problems ?.')



 



import unittest, time
from kcapi import RestURL 

class Testing_URL_API(unittest.TestCase):

    def testing_only_setting_hostname(self): 
        try:
            my_url = RestURL('my-host')
        except Exception as E:
            self.assertEqual("schema" in str(E), True)


        my_url = RestURL('https://my-host')
        self.assertEqual('https://my-host', str(my_url))
        _url = my_url.copy()
        self.assertEqual('https://my-host', str(_url))

    def testing_obj_construction(self):
        self.assertRaises(Exception, lambda: RestURL()) 
        nurl = RestURL(self.endpoint)
        self.assertEqual(str(nurl), self.endpoint)

    def testing_adding_resources(self):
        url = self.url.copy()
        url.addResource("auth", "admin")
        test1 = self.endpoint + "/auth/admin"
        test2 = self.endpoint + "/auth/admin/realms/AAA/client-templates/my_client/scope-mappings/clients/123"
        test3 = self.endpoint + "/auth/admin/realms/AAA/client-templates/my_client_1/scope-mappings/clients/542"

        self.assertEqual(str(url), test1)
        url.addResource("realms", "AAA")

        t1 = url.copy()
        t2 = url.copy()

        t1.addResource("client-templates", "my_client")
        t2.addResource("client-templates", "my_client_1")

        t1.addResource("scope-mappings", "clients")
        t2.addResource("scope-mappings", "clients")

        t1.addResource("clients", "123")
        t2.addResource("clients", "542")

        self.assertEqual(str(t1), test2)
        self.assertEqual(str(t2), test3)

    def testing_single_entry_resources(self):
        url = self.url.copy()

        url.addResources([])
        self.assertEqual(str(url), self.endpoint)

        url.addResources(["auth"])
        self.assertEqual(str(url), self.endpoint + "/auth")
    
    def testing_multiple_entries(self):
        url = self.url.copy()
        raw_str = "/auth/admin/realms/aaa/client-templates/my_client/scope-mappings/clients/123"
        test = raw_str.split("/")
        test =  list( filter(lambda n: n != "", test) )
        url.addResources(test)
        self.assertEqual(str(url), self.endpoint + raw_str)


    def testing_resource_replacement(self):
        url = self.url.copy()
        raw_str = "/auth/admin/realms/aaa/roles"
        test = raw_str.split("/")
        test =  list( filter(lambda n: n != "", test) )
        url.addResources(test)
        self.assertEqual(str(url), self.endpoint + raw_str)

        url.replaceResource('roles', 'roles-by-id')
        self.assertEqual(str(url), self.endpoint + "/auth/admin/realms/aaa/roles-by-id")


    def testing_getting_actual_resource(self):
        self.assertEqual(self.url.getCurrentResource(), "resource1")

    def testing_getting_target(self):
        self.assertEqual(self.url.target(), "https://my-sso.com")

    def testing_replace_current_resource_target(self):
        url = self.url.copy()
        url.replaceCurrentResourceTarget('resource-new')
        self.assertEqual(str(url), "https://my-sso.com/resource-new")


    def testing_removing_resources(self):
        url = "https://my-sso.com"
        myURL = self.url.copy()
        myURL.removeResources(["resource1", "res2"])
        self.assertEqual(str(myURL), url)

    def testing_adding_none_values_to_resources(self):
        urlString = "https://my-sso.com"
        url = RestURL(urlString)
        url.addResource(None, None)
        url.addResources([None, None, None])

        self.assertEqual(urlString, str(url))

        url.addResources([None, None, None, "id", "5"])
        self.assertEqual("https://my-sso.com/id/5", str(url))

    def testing_with_ipv4_host(self):
        url = "http://192.168.1.100:8080/"
        myURL = RestURL(url)
        self.assertEqual(str(myURL), url[:-1])

    def testing_remove_last(self):
        urlString = "https://my-sso.com/a/b/c"
        url = RestURL(urlString)
        url.removeLast()
        self.assertEqual(str(url), "https://my-sso.com/a/b")

    @classmethod
    def setUpClass(self):
        self.endpoint = "https://my-sso.com/resource1"
        self.url = RestURL(self.endpoint)

if __name__ == '__main__':
    unittest.main()



from kcapi import OpenID, Keycloak
import os

def readFromJSON(filename):
    with open(filename) as json_file:
        return json.load(json_file)

class TestBed:
    def __init__(self, realm = None, username = None, password = None, endpoint = None): 

        self.USER = os.getenv('KC_USER')
        self.PASSWORD = os.environ.get('KC_PASSWORD')
        self.REALM = os.environ.get('KC_REALM')
        self.ENDPOINT = os.environ.get('KC_ENDPOINT')

        self.groupName = 'DC'
        self.roleNames = ['level-1', 'level-2', 'level-3'] 
        
        token = OpenID.createAdminClient(self.USER, self.PASSWORD, url=self.ENDPOINT).getToken()
        self.kc = Keycloak(token, self.ENDPOINT)
        self.master_realm = self.kc.admin()
        self.realm = self.REALM 
        self.token = token

    def createRealms(self):
        realm = self.realm
        if self.master_realm.existByKV("id", realm): 
            self.master_realm.removeFirstByKV("id", realm)

        self.master_realm.create({"enabled": "true", "id": realm, "realm": realm})

    def createGroups(self):
        group = self.kc.build('groups', self.realm)
        g_creation_state = group.create({"name": self.groupName}).isOk()
        self.createRoles()


    def createRoles(self):
        roles = self.kc.build('roles', self.realm)
        for role in self.roleNames:
            roles.create({"name": role}).isOk() 


    def createClients(self):
        realm = self.realm
        client = {"enabled":True,
                  "attributes":{},
                  "redirectUris":[],
                  "clientId":"dc",
                  "protocol":"openid-connect", 
                  "directAccessGrantsEnabled":True
                  }

        clients = self.kc.build('clients', realm)
        if not clients.create(client).isOk(): 
            raise Exception('Cannot create Client')

    def createUsers(self):
        realm = self.realm
        test_users = [
                {"enabled":'true',"attributes":{},"username":"batman","firstName":"Bruce", "lastName":"Wayne", "emailVerified":""}, 
                {"enabled":'true',"attributes":{},"username":"superman","firstName":"Clark", "lastName":"Kent", "emailVerified":""}, 
                {"enabled":'true',"attributes":{},"username":"aquaman","firstName":"AAA%", "lastName":"Corrupt", "emailVerified":""}
        ]

        users = self.kc.build('users', realm)
        
        for usr in test_users: 
            users.create(usr).isOk()

    def goodBye(self): 
        state = self.master_realm.remove(self.realm).ok()
        if not state:
            raise Exception("Cannot delete the realm -> " + self.realm )

    def getKeycloak(self):
        return self.kc
    def getAdminRealm(self):
        return self.master_realm

    



from .crud import KeycloakCRUD
from .targets import Targets
from .recovery import add_failiure_recovery_decorator

def AuthenticationFlowURLBuilder(url):
    return Targets.makeWithURL(url).addResources(['flows'])

def add_remote_id_to_payload(kc, payload):
    child_nodes = kc.all()
    for child in child_nodes:
        if child['displayName'] == payload['displayName'] and child['index'] == payload['index'] and child['level'] == \
                payload['level']:
            payload['id'] = child['id']
            return payload


def adjustments_only_update(kc):
    crud_update = kc.update
    return lambda object_id, payload: crud_update(None, add_remote_id_to_payload(kc, payload))


def update_by_doing_delete_and_create(kc, payload):
    unique_identifier = payload['alias'] if 'alias' in payload else payload['displayName']
    key = 'alias' if 'alias' in payload else 'displayName'

    kc.removeFirstByKV(key, unique_identifier)
    kc.create(payload).isOk()
    return True


# These flows cannot be deleted, this mean the user can only adjust parameters.
def is_flow_built_in(root_node):
    kc_default_not_removable_flows = ['browser', 'direct grant', 'registration', 'reset credentials', 'clients',
                                      'first broker login', 'docker auth', 'http challenge']

    return root_node['alias'] in kc_default_not_removable_flows

def choose_update_strategy(kc, root_node):

    if is_flow_built_in(root_node):
        return adjustments_only_update(kc)
    else:
        return kc.update


# The keycloak API is a bit crazy here they add with:
# Post /parentId/executions/execution 
# 
# But they delete with: 
#
# DELETE /executions/<id>
#
# Sadly we need to customize the URL's in order to make it work.
#

def build_action(kc, root_node, action_type):
    parent = root_node['alias']
    resources = [parent, 'executions']

    t = kc.targets.targets
    t['create'].addResources(resources + [action_type])
    t['update'].addResources(resources)
    t['read'].addResources(resources)
    t['delete'].replaceResource('flows', 'executions')

    # Sadly the REST Endpoint to update executions components inside an authentication flow is somewhat broken.
    kc.update = choose_update_strategy(kc, root_node)

    return kc


class AuthenticationFlows(KeycloakCRUD):
    def setURL(self, url):
        super().setURL(url)
        self.targets.addResources(['flows'])

    def _load(self, token, targets):
        flow = KeycloakCRUD()
        flow.token = token
        flow.targets = targets.copy()
        return add_failiure_recovery_decorator(flow)

    def is_built_in(self, root_node):
        return is_flow_built_in(root_node)


    # Generate a CRUD object pointing to /realm/<realm>/authentication/flow_alias/executions/flow
    def flows(self, auth_flow):
        flow = self._load(self.token, self.targets)

        return build_action(
            kc=flow,
            root_node=auth_flow,
            action_type='flow')

    # Generate a CRUD object pointing to /realm/<realm>/authentication/flow_alias/executions/execution
    def executions(self, execution):
        flow = self._load(self.token, self.targets)

        return build_action(
            kc=flow,
            root_node=execution,
            action_type='execution')



def create_flow(flow):
    alias = flow['displayName']
    pid = None if not 'provider' in flow else flow['providerId']
    provider = 'registration-page-flow' if not pid else pid
    flow_type = 'basic-flow' if not pid else 'form-flow'

    # WARN: The value description is not well validated in Keycloak, it can return 500.
    return {
        "alias": alias,
        "type": flow_type,
        "description": "empty",
        "provider": provider,
    }


def get_executions(execution):
    provider = execution['providerId']
    return {'provider': provider}


def is_auth_flow(body):
    return 'authenticationFlow' in body


def get_node_level(node):
    return node['level']


def consistent(parent, flow):
    published_flows = parent.all()

    for publishedFlow in published_flows:
        resource_id = False

        if 'providerId' in publishedFlow and 'providerId' in flow:
            resource_id = publishedFlow['providerId'] == flow['providerId']
        else:
            resource_id = publishedFlow['displayName'] == flow['displayName']

        level_is_equal = publishedFlow['level'] == flow['level']
        index_is_equal = publishedFlow['index'] == flow['index']

        if resource_id and level_is_equal and index_is_equal:
            publishedFlow['requirement'] = flow['requirement']
            parent.update(obj_id=None, payload=publishedFlow).isOk()
            return True

    return False


class AuthenticationFlowsImporter():
    def __init__(self, authentication_api):
        self.flowAPI = authentication_api

    def update(self, root_node, flows):
        if self.flowAPI.is_built_in(root_node):
            for flow in flows:
                if is_auth_flow(flow):
                    self.flowAPI.flows(root_node).update(None, flow).isOk()
                else:
                    self.flowAPI.executions(root_node).update(None, flow).isOk()
        else:
            self.remove(root_node)
            self.flowAPI.create(root_node).isOk()
            self.publish(root_node, flows)

    def remove(self, root_node):
        unique_identifier = root_node['alias'] if 'alias' in root_node else root_node['displayName']
        key = 'alias' if 'alias' in root_node else 'displayName'

        self.flowAPI.removeFirstByKV(key, unique_identifier)

    def publish(self, root_node, flows):
        if not isinstance(flows, list):
            raise Exception("Bad Parameters for Authentication Flow: auth_flow parameter should be an array.")

        root_node = root_node
        nodes = {0: root_node}

        root_flow = self.flowAPI.executions(root_node)

        for flow in flows:
            current_level = get_node_level(flow)
            parent = nodes[current_level]

            if is_auth_flow(flow):
                authentication_flow = create_flow(flow)
                nodes[current_level + 1] = authentication_flow
                self.flowAPI.flows(parent).create(authentication_flow).isOk()
            else:
                execution = get_executions(flow)
                self.flowAPI.executions(parent).create(execution).isOk()

            if not consistent(root_flow, flow):
                raise Exception(
                    'There is an inconsistency problem: Changes are not taking place in the server, latency problems ?.')



from .url import RestURL

''' 
        The guys at Keycloak don't use standard intuitive Rest path for all the resources:

        for example to create a car you can do:
            Post /car, 

        But to delete it you have to do:
            Delete /house/1,  # this delete de car above  

        Thats why we need this class.  
    '''
class Targets: 
    targets = {}

    @staticmethod
    def makeWithURL(url = None): 
        url = RestURL(url)
        t = Targets()

        t.targets = {
            "create": url.copy(),
            "update": url.copy(),
            "delete": url.copy(),
            "read":   url.copy(),
        }

        return t

    @staticmethod
    def makeWithArrayOfURLs(urls = {}): 
        t = Targets()
        t.targets = urls

        return t

    def __init__(self):
        self.targets = {}
        self.root_target = None


    def change(self, resourceName): 
        for method in self.targets:
            self.targets[method].replaceCurrentResourceTarget(resourceName)

    def removeResources(self, resources):
        for method in self.targets:
            self.targets[method].removeResources(resources)

        return self

    def removeLast(self):
        for method in self.targets:
            self.targets[method].removeLast()

        return self

    def addResourcesFor(self, name, resources):
        self.targets[name].addResources(resources)
        return self

    def addResources(self, resources): 
        for method in self.targets:
            self.targets[method].addResources(resources)
        return self

    def copy(self):
        urls = {}
        for key in self.targets:
            urls[key] = self.targets[key].copy()
        return Targets.makeWithArrayOfURLs(urls) 

    def methods(self):
        return self.targets.keys()

    def url(self, method_name):
        if method_name not in self.targets:
            raise Exception('Method: '+ method_name +' not supported')

        return self.targets[method_name].copy()





import json
import requests

from .crud import KeycloakCRUD
from .resp import ResponseHandler


class GroupAndRolesMappingBuilder():
    roles = None
    def build(self, groupName, groups): 
        token = groups.token
        targets = groups.targets

        self.roles = self._load(token, targets, KeycloakCRUD)
        self.rolesMappings = self._load(token, targets, KeycloakCRUD)

        self.rolesMappings = self._initGroupMappingAPI(groupName, self.rolesMappings) 
        self.roles.targets.change('roles')

        self.rolesMappings.add = self.add
        self.rolesMappings.__fetchRoles = self.__fetchRoles
        self.rolesMappings.remove = self.remove

        return self.rolesMappings

    
    def _load(self, token, targets, API):
        api = API()
        api.token = token
        api.targets = targets.copy()

        return api

    def _initGroupMappingAPI(self, groupName, groupAPI):
        groupID = groupAPI.findFirst(groupName)['id']
        groupAPI.targets.addResources([groupID, 'role-mappings', 'realm'])
        return groupAPI

    def __fetchRoles(self, roles): 
        find = self.roles.findFirstByKV
        list_of_roles = list( map(lambda name: find('name', name), roles) )
        for role in list_of_roles:
            if not role:
                raise Exception("One or more roles from the provided list: "+str(roles)+" do not exist.")

        return list_of_roles



    def add(self, roles): 
        populatedListOfRoles = self.__fetchRoles(roles)
        return self.rolesMappings.create(populatedListOfRoles)

    # Another example of Keycloak not using standard REST behaviour.
    # Overriding remove from CRUD
    def remove(self, roles): 
        populatedListOfRoles = self.__fetchRoles(roles)
        remove_target = self.rolesMappings.targets.url('delete')
        headers = self.rolesMappings.headers()

        ret = requests.delete(remove_target, data=json.dumps(populatedListOfRoles), headers=headers )
        return ResponseHandler(remove_target).handleResponse(ret)

class Groups(KeycloakCRUD): 
    def realmRoles(self, group):
        return GroupAndRolesMappingBuilder().build(group, self)






import unittest, time
from kcapi import OpenID, RestURL
from .testbed import TestBed


class Testing_User_API(unittest.TestCase):

    def test_adding_credentials_with_wrong_params(self):
        users = self.testbed.getKeycloak().build('users', self.REALM)
        user_info = {'key': 'username', 'value': 'batman'}
        user_credentials = {'temporary': False, 'passwordWrongParam': '12345'}
        try:
            state = users.updateCredentials(user_info, user_credentials).isOk()
        except Exception as E:
            self.assertEqual("Missing parameter: value" in str(E), True)

    def test_adding_credentials_to_user(self):

        users = self.testbed.getKeycloak().build('users', self.REALM)
        user_info = {'key': 'username', 'value': 'batman'}
        user_credentials = {'temporary': False, 'value': '12345'}
        state = users.updateCredentials(user_info, user_credentials).isOk()
        self.assertTrue(state)

        oid_client = OpenID({
            "client_id": "dc",
            "username": "batman",
            "password": "12345",
            "grant_type": "password",
            "realm": self.REALM
        }, self.testbed.ENDPOINT)

        token = oid_client.getToken()
        self.assertNotEqual(token, None)

    @classmethod
    def setUpClass(self):
        self.testbed = TestBed()
        self.testbed.createRealms()
        self.testbed.createUsers()
        self.testbed.createClients()
        self.REALM = self.testbed.REALM

    @classmethod
    def tearDownClass(self):
        self.testbed.goodBye()


if __name__ == '__main__':
    unittest.main()



import unittest, time
from .testbed import TestBed 

ADMIN_USER = "admin"
ADMIN_PSW  = "admin1234"
REALM = "test_heroes_test"
ENDPOINT = 'https://sso-cvaldezr-stage.apps.sandbox-m2.ll9k.p1.openshiftapps.com'


TEST_REALM = "TESTING"

class Testing_Roles_And_Groups_API(unittest.TestCase):
  
    def testing_roles_creation(self):
        role = {"name": "magic"}
        roles = self.kc.build("roles", self.realm)
        state = roles.create(role).isOk() 
        self.assertTrue(state)

        ret = roles.findFirstByKV('name', 'magic')

        self.assertTrue(ret, "We should get the created role back.")


    def testing_roles_removal(self):
        roles = self.kc.build("roles", self.realm)

        self.assertTrue(roles.removeFirstByKV("name", "magic"))
        ret = roles.findFirstByKV('name', 'magic')
        
        self.assertEqual(len(ret), 0, 'The role should be deleted')


    def testing_group_API(self):
        group_payload = {"name":"my_group"}
        groups = self.kc.build("groups", self.realm)
        state = groups.create(group_payload).isOk()
        self.assertEqual(state, True)

        ret = groups.findFirstByKV('name', 'my_group')
        self.assertTrue(ret, "We should get the created group (my_group) back.")


    def testing_adding_roles_to_group(self):
        groups = self.kc.build('groups', self.realm)
        self.assertTrue(hasattr(groups, "realmRoles"))

        #TestBed class will create one group called "DC"
        #And three roles called [level-1, level-2, level-3]
        group = {"key":"name", "value": self.groupName}
        roles_mapping = groups.realmRoles(group)
        state = roles_mapping.add(self.roleNames)
        self.assertEqual(state.status, 204, 'We should get a HTTP 204 response from the server')


    def testing_handling_adding_non_existing_roles_to_group(self):
        groups = self.kc.build('groups', self.realm)
        self.assertTrue(hasattr(groups, "realmRoles"))

        # TestBed class will create one group called "DC"
        # And three roles called [level-1, level-2, level-3]
        group = {"key": "name", "value": self.groupName}
        roles_mapping = groups.realmRoles(group)
        fail = False
        error = ""
        try:
            state = roles_mapping.add(['do-not-exist-1'])
        except Exception as E:
            fail = True
            error = str(E)

        self.assertTrue(fail, "We raise an role not found error.")
        self.assertEqual(error, "One or more roles from the provided list: ['do-not-exist-1'] do not exist.")

    @classmethod
    def setUpClass(self):
        self.testbed = TestBed(REALM, ADMIN_USER, ADMIN_PSW, ENDPOINT)
        self.testbed.createRealms()
        self.testbed.createGroups()
        self.kc = self.testbed.getKeycloak()
        self.realm = self.testbed.REALM 
        self.master_realm = self.testbed.getAdminRealm()
        self.roleNames = self.testbed.roleNames
        self.groupName = self.testbed.groupName
        
    @classmethod
    def tearDownClass(self):

        if self.master_realm.exist(TEST_REALM): 
            self.master_realm.remove(TEST_REALM)
            return True
        return True

if __name__ == '__main__':
    unittest.main()



import unittest, time
from kcapi.ie import AuthenticationFlowsImporter
from .testbed import TestBed
import json


def create_testing_flows(flows, authenticationFlow):
    for flow in flows:
        state = authenticationFlow.create(flow)


def new_flow_definition(name):
    flow = {"alias": name, "type": "basic-flow", "description": "testing", "provider": "registration-page-form"}
    return dict(flow)


def load_sample(fname):
    f = open(fname)
    file1 = json.loads(f.read())
    f.close()
    return file1


def isConsistent(assertEqual, execs, flows):
    for index in range(len(flows)):
        if 'providerId' in flows[index] and 'providerId' in execs[index]:
            assertEqual(flows[index]['providerId'], execs[index]['providerId'])
        else:
            assertEqual(flows[index]['displayName'], execs[index]['displayName'])

        assertEqual(flows[index]['index'], execs[index]['index'])
        assertEqual(flows[index]['level'], execs[index]['level'])
        assertEqual(flows[index]['requirement'], execs[index]['requirement'])


class TestingAuthenticationFlowsAPI(unittest.TestCase):
    def test_flow_api_instantiation(self):
        flows = self.authenticationFlow.all()
        self.assertTrue(len(flows) > 0)

    def testing_create_a_new_flow(self):
        basic_flow = {
            "alias": "basic",
            "providerId": "basic-flow",
            "description": "my_new_basic_flow",
            "topLevel": True,
            "builtIn": False
        }

        client_flow = {
            "alias": "client",
            "providerId": "client-flow",
            "description": "my_new_client_flow",
            "topLevel": True,
            "builtIn": False
        }

        flows = [basic_flow, client_flow]

        for flow in flows:
            state = self.authenticationFlow.create(flow)
            self.assertTrue(state)

            rows = self.authenticationFlow.findFirst({"key": "alias", "value": flow['alias']})
            self.assertEqual(rows['alias'], flow['alias'])
            self.assertEqual(rows['providerId'], flow['providerId'])

    def testing_nested_flows(self):
        basic_flow = self.flows[0]
        name = '--yyyyyyy---'
        nestedFlow = new_flow_definition(name)
        nestedExecution = {"provider": "auth-x509-client-username-form"}

        flows = self.authenticationFlow.flows(basic_flow)
        resp = flows.create(nestedFlow)
        self.assertTrue(resp.isOk())

        nested_flows = flows.all()
        self.assertTrue(len(nested_flows) > 0)

        flw = nested_flows[0]
        self.assertEqual(nestedFlow["alias"], flw["displayName"])

        executions = self.authenticationFlow.executions(basic_flow)
        executions.create(nestedExecution)
        executions.create(nestedExecution)

        x509 = executions.findFirstByKV('displayName', 'X509/Validate Username Form')

        self.assertIsNotNone(x509)
        flow_size = len(flows.all())
        self.assertEqual(flow_size, 3)

        execs_size = len(executions.all())
        self.assertEqual(execs_size, 3)

    def testing_remove_executions_flows(self):
        client_flow = self.flows[1]
        nestedExecution = {"provider": "client-x509"}
        execs = self.authenticationFlow.executions(client_flow)

        resp = execs.create(nestedExecution)
        self.assertTrue(resp.isOk())

        resp = execs.create(nestedExecution)
        self.assertTrue(resp.isOk())

        elen = len(execs.all())
        self.assertEqual(elen, 2)

        execs.removeFirstByKV('providerId', 'client-x509')

        elen = len(execs.all())
        self.assertEqual(elen, 1)

        execs.removeFirstByKV('providerId', 'client-x509')

        elen = len(execs.all())
        self.assertEqual(elen, 0)

    def testing_remove_nested_flows(self):
        basic_flow = self.flows[2]
        nf1 = {"alias": "_aaaaaa_", "type": "basic-flow", "description": "11111111",
               "provider": "registration-page-form"}

        nf2 = {"alias": "_bbbbbb_", "type": "basic-flow", "description": "22222222",
               "provider": "registration-page-form"}

        flows = self.authenticationFlow.flows(basic_flow)
        state1 = flows.create(nf1)
        state2 = flows.create(nf2)

        self.assertTrue((state1 and state2))

        flows_list = flows.all()

        self.assertTrue(len(flows_list) == 2)
        flows.removeFirstByKV('displayName', nf1['alias'])
        flows_list = flows.all()
        self.assertTrue(len(flows_list) == 1)
        self.assertEqual(flows_list[0]['displayName'], nf2['alias'])

    def testing_flows_inside_flows(self):
        parent_flow = self.flows[3]
        x1_def = new_flow_definition('x1')
        x12_def = new_flow_definition('x12')
        x123_def = new_flow_definition('x123')

        # Top node
        # parent --> child x1
        parent = self.authenticationFlow.flows(parent_flow)
        state = parent.create(x1_def).isOk()
        self.assertTrue(state)

        # Adding child
        # parent --> child x1 --> x12
        x1 = self.authenticationFlow.flows(x1_def)
        state = x1.create(x12_def).isOk()
        self.assertTrue(state)

        # Adding nested child
        # parent --> child x1 --> x12 --> x123
        x12 = self.authenticationFlow.flows(x12_def)
        state = x12.create(x123_def).isOk()
        self.assertTrue(state)

        # Adding a nested executor to the last node
        # parent --> child x1 --> x12 --> x123
        #                                 | 
        #                                 v 
        #                             client-x509
        x123 = self.authenticationFlow.executions(x123_def)
        execution = {"provider": "docker-http-basic-authenticator"}
        state = x123.create(execution).isOk()
        self.assertTrue(state)

        executions = self.authenticationFlow.executions(parent_flow)
        exec_size = len(parent.all())

        self.assertEqual(exec_size, 4)

    def testing_import_authentication_flows(self):
        parent_flow = self.flows[6]
        publisher = AuthenticationFlowsImporter(self.authenticationFlow)
        self.assertIsNotNone(publisher)

        flows = load_sample('./test/payloads/nested.json')

        publisher.publish(parent_flow, flows)

        execs = self.authenticationFlow.executions(parent_flow).all()
        isConsistent(self.assertEqual, flows, execs)

    def testing_import_authentication_flows_2(self):
        parent_flow = self.flows[5]
        publisher = AuthenticationFlowsImporter(self.authenticationFlow)
        self.assertIsNotNone(publisher)

        flows = load_sample('./test/payloads/nested_2.json')

        publisher.publish(parent_flow, flows)

        execs = self.authenticationFlow.executions(parent_flow).all()
        isConsistent(self.assertEqual, flows, execs)

    def testing_import_authentication_flows_1(self):
        parent_flow = self.flows[4]
        publisher = AuthenticationFlowsImporter(self.authenticationFlow)
        self.assertIsNotNone(publisher)

        flows = load_sample('./test/payloads/nested_1.json')

        publisher.publish(parent_flow, flows)

        execs = self.authenticationFlow.executions(parent_flow).all()
        isConsistent(self.assertEqual, flows, execs)

    def testing_performing_authentication_built_in_flows_updates(self):
        publisher = AuthenticationFlowsImporter(self.authenticationFlow)

        clients_flow = {
            "alias": "clients"
        }

        executors = load_sample('./test/payloads/executors/executor_1.json')

        publisher.update(clients_flow, executors)
        execs = self.authenticationFlow.executions(clients_flow).all()

        self.assertEqual(execs[0]['requirement'], 'REQUIRED')
        self.assertEqual(execs[3]['requirement'], 'REQUIRED')

        reset_credentials_flow = {
            "alias": "reset credentials"
        }

        reset_credentials_flow_payload = load_sample('./test/payloads/executors/executor_2.json')

        publisher.update(reset_credentials_flow, reset_credentials_flow_payload)
        execs2 = self.authenticationFlow.executions(reset_credentials_flow).all()

    def testing_performing_authentication_flows_updates(self):
        publisher = AuthenticationFlowsImporter(self.authenticationFlow)

        update_flow = {
            "alias": "update_flow_1",
            "providerId": "client-flow",
            "description": "my_new_client_flow",
            "topLevel": True,
            "builtIn": False
        }

        self.authenticationFlow.create(update_flow)

        flows = load_sample('./test/payloads/executors/flow_1.json')
        publisher.update(update_flow, flows)

        amended_flow = self.authenticationFlow.executions(update_flow).all()
        isConsistent(self.assertEqual, flows, amended_flow)


    @classmethod
    def setUpClass(self):
        self.testbed = TestBed()
        self.testbed.createRealms()
        self.testbed.createUsers()
        self.testbed.createClients()
        self.authenticationFlow = self.testbed.getKeycloak().build('authentication', self.testbed.REALM)

        basic_flow = {
            "alias": "test_number_0",
            "providerId": "basic-flow",
            "description": "my_new_basic_flow",
            "topLevel": True,
            "builtIn": False
        }

        client_flow = {
            "alias": "test_number_1",
            "providerId": "client-flow",
            "description": "my_new_client_flow",
            "topLevel": True,
            "builtIn": False
        }

        basic_flow_2 = {
            "alias": "test_number_2",
            "providerId": "basic-flow",
            "description": "my_new_basic_flow",
            "topLevel": True,
            "builtIn": False
        }

        flow_3 = {
            "alias": "test_number_3",
            "providerId": "basic-flow",
            "description": "my_new_basic_flow",
            "topLevel": True,
            "builtIn": False
        }

        flow_4 = {
            "alias": "test_number_4",
            "providerId": "basic-flow",
            "description": "my_new_basic_flow",
            "topLevel": True,
            "builtIn": False
        }

        flow_5 = {
            "alias": "test_number_5",
            "providerId": "basic-flow",
            "description": "my_new_basic_flow",
            "topLevel": True,
            "builtIn": False
        }

        flow_6 = {
            "alias": "test_number_6",
            "providerId": "basic-flow",
            "description": "my_new_basic_flow",
            "topLevel": True,
            "builtIn": False
        }

        self.flows = [basic_flow, client_flow, basic_flow_2, flow_3, flow_4, flow_5, flow_6]
        create_testing_flows(self.flows, self.authenticationFlow)

    @classmethod
    def tearDownClass(self):
        # self.testbed.goodBye()
        return True


if __name__ == '__main__':
    unittest.main()



from .crud import KeycloakCRUD
from .targets import Targets


def RealmURLBuilder(url):
    return Targets.makeWithURL(url)


class KeycloakCaches: 
    def __init__(self, kcCRUD, realmName):
        self.name = realmName 
        self.request_body = {'realm': self.name}
        self.kcrud = KeycloakCRUD()
        self.kcrud.token = kcCRUD.token
        self.kcrud.targets = kcCRUD.targets.copy().removeLast()

    def postTo(self, target):
        createRef =self.kcrud.targets.targets['create']
        createRef.addResource(target)
        
        ret = self.kcrud.create(self.request_body)
        
        createRef.removeLast()
        return ret

    def clearUserCache(self):
        return self.postTo('clear-realm-cache')

    def clearRealmCache(self): 
        return self.postTo('clear-user-cache')

    def clearKeyCache(self):
        return self.postTo('clear-keys-cache')

class Realms(KeycloakCRUD):
    def caches(self, realmName): 
        return KeycloakCaches(self, realmName)



import unittest, time
from kcapi import OpenID, Keycloak
from .testbed import TestBed 

ADMIN_USER = "admin"
ADMIN_PSW  = "admin1234"
REALM = "test_heroes_test"
ENDPOINT = 'https://sso-cvaldezr-stage.apps.sandbox-m2.ll9k.p1.openshiftapps.com'


TEST_REALM = "TESTING"

class Testing_Mappings(unittest.TestCase):
  
    def testing_adding_roles_to_group(self):
        groups = self.kc.build('groups', self.realm)
        self.assertTrue(hasattr(groups, "realmRoles"))

        #TestBed class will create one group called "DC"
        #And three roles called [level-1, level-2, level-3]

        group = {"key":"name", "value": self.groupName}
        roles_mapping = groups.realmRoles(group)
        state = roles_mapping.add(self.roleNames)
        self.assertTrue(state)

    def testing_all_roles_to_group(self):    
        groups = self.kc.build('groups', self.realm)
        group = {"key":"name", "value": self.groupName}
        roles_mapping = groups.realmRoles(group)
        rolesInGroup = roles_mapping.all()

        self.assertEqual(len(rolesInGroup), 3, 'should be 3 roles: [level-1, level-2, level-3]')

    def testing_remove_roles_from_group(self):    
        groups = self.kc.build('groups', self.realm)
        group = {"key":"name", "value": self.groupName}
        roles_mapping = groups.realmRoles(group)

        roles_mapping.remove(["level-1", "level-2"])

        rolesInGroup = roles_mapping.all()

        self.assertEqual(len(rolesInGroup), 1, 'we should got only 1 role: [level-3]')

    def testing_adding_user_to_group(self):
        users = self.kc.build('users', self.realm)
        self.assertTrue(hasattr(users, "joinGroup"))

        usr = {"key": "username", "value": "batman"}
        gpr = {"key": "name", "value": self.groupName}

        join_state = users.joinGroup(usr, gpr).isOk()
        self.assertTrue(join_state)

    def testing_user_leaving_group(self):
        users = self.kc.build('users', self.realm)
        self.assertTrue(hasattr(users, "joinGroup"))

        usr = {"key": "username", "value": "batman"}
        gpr = {"key": "name", "value": self.groupName}

        leave_status = users.leaveGroup(usr, gpr).isOk()
        self.assertTrue(leave_status)

        groups = users.groups(usr)
        self.assertEqual(len(groups), 0)

    def testing_user_group_mapping(self):
        users = self.kc.build('users', self.realm)
        usr = {"key": "username", "value": "batman"}
        batman = users.groupMapping(usr)

        self.assertEqual(len(batman.all()), 1, 'the user batman should belong to one group')
 

    @classmethod
    def setUpClass(self):
        self.testbed = TestBed(REALM, ADMIN_USER, ADMIN_PSW, ENDPOINT)
        self.testbed.createRealms()
        self.testbed.createUsers()
        self.testbed.createGroups()
        self.kc = self.testbed.getKeycloak()
        self.realm = self.testbed.REALM 
        self.master_realm = self.testbed.getAdminRealm()
        self.roleNames = self.testbed.roleNames
        self.groupName = self.testbed.groupName
        
    @classmethod
    def tearDownClass(self):
        #self.testbed.goodBye()
        if self.master_realm.exist(TEST_REALM): 
            #self.master_realm.remove(TEST_REALM)
            return True
        return True

if __name__ == '__main__':
    unittest.main()



from .targets import Targets


# The Keycloak guys decided to use another resource DELETE /roles-by-id, instead of sticking to DELETE /roles.
def RolesURLBuilder(url):
    targets = Targets.makeWithURL(url)
    targets.targets['delete'].replaceResource('roles', 'roles-by-id')
    return targets


import unittest, time
from .testbed import TestBed
import json


def load_sample(fname):
    f = open(fname)
    file1 = json.loads(f.read())
    f.close()
    return file1


class TestingUserAPI(unittest.TestCase):

    def test_identity_provider_adding_saml_and_oid_IDP_providers(self):
        idp1 = self.kc.build('identity-provider', self.REALM)
        idp = self.kc.build('idp', self.REALM)
        saml = load_sample('./test/payloads/idp_saml.json')
        oid = load_sample('./test/payloads/idp_oid.json')
        state1 = idp.create(saml).isOk()
        self.assertTrue(state1)

        state2 = idp1.create(oid).isOk()
        self.assertTrue(state2)

        resources = idp1.all()

        self.assertEqual(resources[0]['alias'], saml['alias'], 'SAML alias should match')
        self.assertEqual(resources[1]['alias'], oid['alias'], 'SAML alias should match')

    @classmethod
    def setUpClass(self):
        self.testbed = TestBed()
        self.testbed.createRealms()
        self.testbed.createUsers()
        self.testbed.createClients()
        self.REALM = self.testbed.REALM
        self.kc = self.testbed.getKeycloak()

    @classmethod
    def tearDownClass(self):
        # self.testbed.goodBye()
        return True


if __name__ == '__main__':
    unittest.main()



DEBUG = False

class ResponseHandler:
    def __handle_failure(self, resp, url):
        code = resp.status_code

        if code in [503, 500, 409, 404, 400]:
            raise Exception('Method: '+ self.method +", Server Error: " + str(code) +  " Message: "+ resp.text +  " URL: ", str(url), ' payload: ', self.payload)

        if code == 401:
            raise Exception("URL: ", url ," Method: ", self.method, ", Server returned 401: Unauthorized. Please check username or password.")


        json_data = resp.json()
        error_message = json_data["error"] + "--" + json_data["error_description"]
        raise Exception("Method: "+ self.method +" Error: " + str(code) + " \n for URL:" + str(url) + " \n Response: " +  error_message)


    def __init__(self, urlObject, method = None, payload=None):
        self.URL = urlObject
        self.method = method
        self.payload = payload

    def handleResponse(self, response):
        self.response = response
        self.status = response.status_code
        return self

    def verify(self):
        if self.isOk():
            return self

    def isOk(self):
        if self.status >= 200 and self.status <=208:
            return True
        else:
            self.__handle_failure(self.response, self.URL)

        return self

    def ok(self):
        return self.status >= 200 and self.status <= 208

    def resp(self):
        return self.response

    def response(self):
        return self.response



from setuptools import setup

with open('README.md') as f:
    readme = f.read()


# This call to setup() does all the work
setup(
    name="kcapi",
    version="1.0.18",
    description="A module to automate stuff in Keycloak.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/cesarvr/keycloak-api",
    author="cesar valdez",
    author_email="cesar.valdez@gosub.biz",
    license="MIT",
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy"
    ],
    packages=["kcapi", "kcapi.rest", "kcapi.ie"],
    include_package_data=False,
    install_requires=["requests"],
)



import requests, json, time
from .rest import RestURL

def craft_error_message(resp, url):
    code = resp.status_code

    if code in [404]:
        raise Exception("Server Error: " + str(code), resp.text, " URL: ", str(url))

    if code in [503, 500]:
        raise Exception("Server Error: " + str(code), " URL: ", str(url))

    if code == 401:
        raise Exception("Server returned 401: Unauthorized. Please check username or password.")

    json_data = resp.json()
    error_message = json_data["error"] + "--" + json_data["error_description"]
    raise Exception("Error: " + str(code) + " \n for URL:" + str(url) + " \n Response: " + error_message)


# Retrieves the Well Known Endpoint: https://openid.net/specs/openid-connect-discovery-1_0.html
def get_well_known_info(url, realm):
    discovery_url = RestURL(url, ['auth', 'realms', realm, '.well-known', 'openid-configuration'])

    resp = requests.get(url=str(discovery_url))

    if resp.status_code == 200:
        return resp.json()

    craft_error_message(resp, url)


def elapsed_time(time2):
    return (time.time() - time2)

class Token:
    def __init__(self, well_known={}, payload=None, refresh_token=None, raw_json_str=None, client_id=None):
        self.token = None
        self.refresh_token = None
        self.expiring = 60
        self.well_known = well_known
        self.start_time = 0
        self.payload = None
        self.client_id = client_id if client_id else "admin-cli"


        if payload:
            self.load_from_request_payload(payload)

        if refresh_token:
            self.load_from_refresh_token(refresh_token)

        if raw_json_str:
            self.load_from_request_payload(json.loads(raw_json_str.replace("'", "\"")))

        if not payload and not refresh_token and not raw_json_str:
            raise Exception('Token not properly initialized. You have to provide a single refresh token, or otherwise a dictionary/raw JSON string with the following shape: https://datatracker.ietf.org/doc/html/rfc6749#content.')


    def load_from_refresh_token(self, refresh_token):
        self.refresh_token = refresh_token
        self = self.refresh()

    def load_from_request_payload(self, payload):
        self.token = payload['access_token']
        self.refresh_token = payload['refresh_token']
        self.expiring = int(payload['expires_in'])
        self.start_time = time.time()
        self.payload = payload


    def expired(self):
        if elapsed_time(self.start_time) >= (self.expiring - 15): # If current time is above expiring time minus 10 seconds we ask for a new token.
            return True

        return False

    def refresh(self):
        token_endpoint = self.well_known['token_endpoint']

        body = {
            "client_id": self.client_id,
            "grant_type": "refresh_token",
            "refresh_token": self.refresh_token
        }

        resp = requests.post(token_endpoint, data=body)

        if resp.status_code != 200:
            craft_error_message(resp, token_endpoint)

        return Token(payload=resp.json(), well_known=self.well_known)

    def get_token(self):
        return self.token

    def __str__(self):
        return str(self.payload)

    def __repr__(self):
        return str(self.payload)

class OpenID:
    def __check_params(self, params):
        expected_params = ['password', 'username', 'grant_type', 'client_id']

        for param in expected_params:
            if param not in params:
                raise Exception("Missing parameter on OpenId class: ", param)


    def __init__(self, credentials, url=None):
        self.__check_params(credentials)

        if not url:
            raise Exception('URL Not Found: Make sure you provide a URL before invoking the service')

        self.credentials = credentials
        self.realm = self.credentials['realm']
        self.token = None
        self.url = url



    @staticmethod
    def createAdminClient(username=None, password=None, url = None):
        __props = {
            "client_id": "admin-cli",
            "grant_type": "password",
            "realm": "master",
            "username": username,
            "password": password
        }

        return OpenID(__props, url)

    def getToken(self):
        url = self.url
        well_known = get_well_known_info(url, self.realm)
        resp = requests.post(well_known['token_endpoint'], data=self.credentials)

        if resp.status_code == 200:
            payload = resp.json()
            return Token(payload=payload, well_known=well_known)
        else:
            craft_error_message(resp, url)



from .rest.crud import KeycloakCRUD
from .rest.targets import Targets
from .rest.groups import Groups
from .rest.roles import RolesURLBuilder
from .rest.users import Users
from .rest.realms import Realms, RealmURLBuilder
from .rest.url import RestURL
from .rest.idp import IdentityProviderURLBuilder
from .rest.auth_flows import AuthenticationFlows, AuthenticationFlowURLBuilder
from .rest.clients import Clients
from .rest.recovery import add_failiure_recovery_decorator

from .oid import get_well_known_info, Token

KCResourceTypes = {
    "users": Users,
    "groups": Groups,
    "realms": Realms,
    "authentication": AuthenticationFlows,
    "clients": Clients
}

URLBuilders = {
    'roles': RolesURLBuilder,
    'authentication': AuthenticationFlowURLBuilder,
    "idp": IdentityProviderURLBuilder,
    "identity-provider": IdentityProviderURLBuilder,
    "realms": RealmURLBuilder,
}


def GenericURLBuilder(url):
    targets = Targets.makeWithURL(url)
    return targets

def resource(resource_name=None, token=None):
    KCResourceAPI = KeycloakCRUD if not resource_name in KCResourceTypes else KCResourceTypes[resource_name]
    resource = KCResourceAPI()
    resource.token = token

    return resource

def configure_url(url=None, resource_name=None):
    URLBuilder = GenericURLBuilder if not resource_name in URLBuilders else URLBuilders[resource_name]
    return URLBuilder(str(url))


MASTER_REALM = 'master'
class Keycloak:
    @staticmethod
    def instantiate_with_refresh_token(url=None, refresh_token=None, client_id=None):
        well_known = get_well_known_info(url,MASTER_REALM)
        token = Token(well_known=well_known, refresh_token=refresh_token, client_id=client_id)
        token.client_id = token.client_id if not client_id else client_id
        return Keycloak(token=token, url=url)

    @staticmethod
    def instantiate_with_raw_json(url=None, json_string=None, client_id=None):
        well_known = get_well_known_info(url, MASTER_REALM)
        token = Token(well_known=well_known, raw_json_str=json_string, client_id=client_id)
        return Keycloak(token=token, url=url)

    def __init__(self, token=None, url=None):
        if not token:
            raise Exception("No authentication token provided.")

        if not url:
            raise Exception("No Keycloak endpoint URL has been provided.")

        self.token = token
        self.url = RestURL(url=url, resources=["auth", "admin", "realms"])

    def admin(self):
        adm = resource(token=self.token)
        adm.targets = configure_url(url=self.url.copy())
        return add_failiure_recovery_decorator(adm)

    def build(self, resource_name, realm):
        url = self.url.copy()
        url.addResources([realm, resource_name])

        res = resource(token=self.token, resource_name=resource_name)
        res.targets = configure_url(url=url, resource_name=resource_name)
        return add_failiure_recovery_decorator(res)



import requests, json
from .resp import ResponseHandler

class KeycloakCRUD(object):
    @staticmethod
    def get_child(that, client_id, resource_name):
        kc = KeycloakCRUD()
        kc.token = that.token
        kc.targets = that.targets.copy()

        kc.targets.addResources([client_id, resource_name])

        return kc

    def __init__(self): 
        self.targets = None
        self.token = None

    def headers(self):

        if self.token.expired():
            self.token = self.token.refresh()

        return {
                'Content-type': 'application/json', 
                'Authorization': 'Bearer '+ self.token.get_token()
        }

    def setIdentifier(self, _id = None, url = None):
        if _id:
            return url.addResource(_id)
        else:
            return url
    
    def create(self, payload):
        url = self.targets.url('create')

        ret = requests.post(url, data=json.dumps(payload), headers=self.headers())
        return ResponseHandler(url, method='Post', payload=payload).handleResponse(ret)

    def update(self, obj_id=None, payload=None):
        url = self.targets.url('update')
        target = str(self.setIdentifier(obj_id, url))

        ret = requests.put(target, data=json.dumps(payload), headers=self.headers())
        return ResponseHandler(target, method='Put', payload=payload).handleResponse(ret)

    def remove(self, _id):
        delete = self.targets.url('delete')
        url = self.setIdentifier(_id, delete)
        ret = requests.delete(url, headers=self.headers())
        return ResponseHandler(url, method='Delete').handleResponse(ret)
        
    def get(self, _id):
        url = self.targets.url('read')
        ret = requests.get(str(self.setIdentifier(_id, url)), headers=self.headers())
        return ResponseHandler(url, method='Get').handleResponse(ret)

    def findAll(self):
        url = self.targets.url('read')
        ret = requests.get(url, headers=self.headers())
        return ResponseHandler(url, method='Get').handleResponse(ret)

    def findFirst(self, params):
        return self.findFirstByKV(params['key'], params['value'])

    def findFirstByKV(self, key, value):
        rows = self.findAll().verify().resp().json()

        for row in rows: 
            if row[key].lower() == value.lower():
                return row

        return []

    def all(self):
        return self.findAll().verify().resp().json()

    def updateUsingKV(self, key, value, obj): 
        res_data = self.findFirstByKV(key,value)

        if res_data: 
            data_id = res_data['id']
            res_data.update(obj)
            return self.update(data_id, res_data).isOk() 
        else:
            return False

    def removeFirstByKV(self, key, value, custom_key="id"):
        row = self.findFirstByKV(key,value)

        if row:
            return self.remove(row[custom_key]).isOk()
        else:
            return False

    def existByKV(self, key, value): 
        ret = self.findFirstByKV(key, value)
        return ret != False

    def exist(self, _id):
        return self.get(_id).ok()


